body {
    min-height: 100vh;
    min-height: -webkit-fill-available;
    font-family: 'Poppins', sans-serif;
    font-weight: lighter;
}

html {
    height: -webkit-fill-available;
}

main {
    display: flex;
    flex-wrap: nowrap;
    height: 100vh;
    height: -webkit-fill-available;
    max-height: 100vh;
    overflow-x: auto;
    overflow-y: hidden;
    background: #f6f6f6;
}

.app-title {
    display: flex;
    justify-content: center;
}

.b-example-divider {
    flex-shrink: 0;
    width: 1.5rem;
    height: 100vh;
}

.bi {
    vertical-align: -.125em;
    pointer-events: none;
    fill: currentColor;
}

.dropdown-toggle {
    outline: 0;
}

.nav-flush .nav-link {
    border-radius: 0;
}

.btn-primary {
    background-color: #009191;
}

.btn-primary:hover {
    background-color: #009999;
}

.btn-primary:focus {
    background-color: #009999;
    box-shadow: 0 0 0 0.25rem rgb(#00999977);
}

.btn-toggle {
    display: inline-flex;
    align-items: center;
    padding: .25rem .5rem;
    color: rgba(0, 0, 0, .65);
    background-color: transparent;
    border: 0;
}

.btn-toggle:hover,
.btn-toggle:focus {
    color: rgba(0, 0, 0, .85);
    background-color: #d2f4ea;
}

.btn-toggle::before {
    width: 1.25em;
    line-height: 0;
    content: url("data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' width='16' height='16' viewBox='0 0 16 16'%3e%3cpath fill='none' stroke='rgba%280,0,0,.5%29' stroke-linecap='round' stroke-linejoin='round' stroke-width='2' d='M5 14l6-6-6-6'/%3e%3c/svg%3e");
    transition: transform .35s ease;
    transform-origin: .5em 50%;
}

.btn-toggle[aria-expanded="true"] {
    color: rgba(0, 0, 0, .85);
}

.btn-toggle[aria-expanded="true"]::before {
    transform: rotate(90deg);
}

.btn-toggle-nav a {
    display: inline-flex;
    padding: .1875rem .5rem;
    margin-top: .125rem;
    margin-left: 1.25rem;
    text-decoration: none;
}

.btn-toggle-nav a:hover,
.btn-toggle-nav a:focus {
    background-color: #d2f4ea;
}

.scrollarea {
    overflow-y: auto;
}

.fw-semibold {
    font-weight: 600;
}

.lh-tight {
    line-height: 1.25;
}

.sidebar {
    width: 20rem;
    background: #FBFCFD;
    border: 1px solid #E5E5E5;
}

footer {
    align-self: center;
    margin-top: auto;
    font-weight: lighter;
}

footer a {
    color: #009191;
    text-decoration: none;
    font-size: 12px;
}

.options-title {
    font-size: 14px;
}

.accordion-item {
    background: unset;
    border: unset;
    margin-top: 2rem;
}

.accordion-button {
    padding: 0;
    background: transparent;
}

.accordion-button:focus {
    z-index: 3;
    border-color: unset;
    outline: 0;
    box-shadow: none;
}

.accordion-button:not(.collapsed) {
    color: black;
    background-color: transparent;
    box-shadow: none;
}

.accordion-button:not(.collapsed)::after {
    background-image: url("data:image/svg+xml,%3csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 16 16' fill='%23212529'%3e%3cpath fill-rule='evenodd' d='M1.646 4.646a.5.5 0 0 1 .708 0L8 10.293l5.646-5.647a.5.5 0 0 1 .708.708l-6 6a.5.5 0 0 1-.708 0l-6-6a.5.5 0 0 1 0-.708z'/%3e%3c/svg%3e");
    transform: rotate(-180deg);
}

.accordion-body {
    padding: 0;
}

.progress {
    border-radius: 0;
    width: 20rem;
    margin-left: -1.1rem;
}

.progress-positive {
    background: #009191;
}

.progress-negative {
    background: #F24E1E;
}

.text-end {
    align-self: center;
}

.col-form-label {
    color: #00000077;
    font-weight: lighter;
    font-size: 14px;
}

.form-control, .form-select {
    border: 1px solid #F2F2F2;
    box-sizing: border-box;
    border-radius: 4px;
    font-size: 14px;
    line-height: 21px;
    letter-spacing: 0.1px;
    color: rgba(0, 0, 0, .85)
}

.form-check-input:checked {
    background-color: #198754;
    border-color: #198754;
}

.form-check-input[type=checkbox] {
    border-radius: .5em;
}

.list-group-item {
    padding: 0;
    border: none;
    border-bottom: 1px solid #E5E5E5;
}

.match-item {
    display: flex;
    justify-content: space-between;
    align-items: center;
    font-weight: lighter;
    font-size: 14px;
    line-height: 21px;
    letter-spacing: 0.1px;
    padding: 1rem;
    width: 100%;
}

.btn {
    display: flex;
    align-items: center;
}

.kbd {
    position: absolute;
    top: 0;
    right: 0;
    line-height: 0.8rem;
    color: #000000;
    background-color: #fff;
    cursor: pointer;

    border: 1px solid gray;
    font-size: 1.2em;
    box-shadow: 1px 0 1px 0 #eee, 0 2px 0 2px #ccc, 0 2px 0 3px #444;
    -webkit-border-radius: 3px;
    -moz-border-radius: 3px;
    border-radius: 3px;
    padding: 6px 2px;
    max-height: 1em;
    font-family: monospace;
    min-width: 1em;
}

.btn .kbd {
    position: relative;
    top: -10px;
    right: -10px;
}

.list-group-item .kbd {
    position: relative;
}

.hint-text {
    max-width: 50%;
}

.text-card .positive-match::before {
    display: none;
}

.positive-match {
    background: #E6F5F5;
    color: #009191;
}

.spinner-busy {
    position: absolute;
    bottom: 0;
    right: 0;
    margin: 50px;
}

.alert-loss-connect {
    position: absolute;
    margin-left: auto;
    margin-right: auto;
}

[v-cloak] {
    display: none;
}

.card-header {
    font-weight: 600;
    font-size: 12px;
    line-height: 21px;
    letter-spacing: 0.1px;
    color: #fff;
    background: #009191;
}

.image-matches-container {
    display: grid;
    gap: .25rem;
    grid-template-columns: repeat(3, 1fr);
    grid-auto-rows: minmax(100px, auto);
}

.card {
    overflow-y: auto;
}

.card-body {
    padding: 0;
}

.image-card .card-body {
    padding: .25rem;
}

.image-card .card-body .positive-match .img-thumbnail {
    background: transparent;
    padding: .75rem;
    border: none;
}

.positive-match:before {
    background: url(./img/check.svg);
    content: '';
    position: absolute;
    height: 1.5rem;
    width: 1.5rem;
    padding: .5rem;
}

.image-card .card-header {
    display: flex;
    justify-content: center;
    align-items: center;
    padding: .5rem;
}

.image-card .card-body .row {
    width: 100%;
}

.image-card .card-body .compact-img {
    position: relative;
    display: flex;
    justify-content: center;
}

.image-card .card-header .img-thumbnail {
    margin-left: .25rem;
    width: 20vh;
    height: 20vh;
    object-fit: contain;
    padding: 0;
}

.image-card .card-body .img-thumbnail {
    width: 20vh;
    height: 20vh;
    object-fit: contain;
    padding: 0;
}

.mesh-card .card-header {
    display: flex;
    justify-content: center;
    align-items: center;
    padding: .5rem;
    height: 10rem;
}

/* Swiper don't swipe */
.swiper-container {
    width: 100%;
    height: 100%;
}

.swiper-slide {
    display: flex;
    justify-content: center;
    align-items: center;
}

.swiper-container-3d .swiper-slide-shadow-right, .swiper-container-3d .swiper-slide-shadow-left {
    background: unset;
}

.swiper-slide-prev > .card, .swiper-slide-next > .card {
    opacity: .5;
    filter: blur(2px) grayscale(1);
    pointer-events: none;
}

:root {
    --swiper-theme-color: #009191;
    --swiper-navigation-size: 20px;
}

#accordionPanelsStayOpenExample {
    overflow-y: scroll;
    overflow-x: hidden;
    height: 100%;
}